# This file is released under the MIT License
# License does not apply to files generated through this script

import os
from pathlib import Path
import json
from difflib import SequenceMatcher, ndiff
import markdown
from unicodedata import normalize
from collections import defaultdict
import sys

import_dir = f"{sys.argv[1].upper()}/{sys.argv[2]}"
playback_dir = f"playback_{sys.argv[1]}"
output_name = sys.argv[2]

footnote_ch = "⁰¹²³⁴⁵⁶⁷⁸⁹"
footnote_ch_full = footnote_ch + "⁽⁾"

class Event:
    seq_num = 0
    events = []
    init_done = False

    def __init__(self):
        self.id = f"ev-{Event.seq_num}"
        self.timestamp = 0
        self.createdByDevGroupId = "devGroupId-1"
        self.eventSequenceNumber = Event.seq_num
        self.permanentRelevance = "never relevant" if not Event.init_done else ""
        Event.seq_num += 1
        Event.events.append(self)

class CreateDirEvent(Event):
    def __init__(self, name):
        super().__init__()
        self.type = "CREATE DIRECTORY"
        if name == "/":
            self.directoryId = "/"
            self.directoryPath = "/"
        else:
            name = "/" + name
            self.directoryId = name + "/"
            self.directoryPath = name + "/"
        parent = Path(name).parent
        if str(parent) != "/" and str(parent) != str(name):
            self.parentDirectoryId = str(parent) + "/"
        else:
            self.parentDirectoryId = None

class CreateFileEvent(Event):
    def __init__(self, name):
        super().__init__()
        name = "/" + name
        self.type = "CREATE FILE"
        self.fileId = name
        self.filePath = name
        parent = Path(name).parent
        if str(parent) == "/":
            self.parentDirectoryId = "/"
        else:
            self.parentDirectoryId = str(parent) + "/"

class InsertFileEvent(Event):
    def __init__(self, name):
        super().__init__()
        self.type = "INSERT"
        self.fileId = "/" + name
        self.character = ""
        self.previousNeighborId = None
        self.lineNumber = 0
        self.column = 0
        self.pastedEventId = None

class DeleteFileEvent(Event):
    def __init__(self, name):
        super().__init__()
        self.type = "DELETE"
        self.fileId = "/" + name
        self.character = ""
        self.previousNeighborId = None
        self.lineNumber = 0
        self.column = 0

class NoopEvent(Event):
    def __init__(self, name):
        super().__init__()
        self.type = "NOOP"
        self.fileId = "/" + name
        self.lineNumber = 0
        self.column = 0

class Comment():
    seq_num = 0
    comments = []

    def __init__(self, evt, sc):
        self.id = f"commentId-{Comment.seq_num}"
        self.timestamp = 0
        self.displayCommentEvent = evt
        self.developerGroupId = "devGroupId-1"
        self.commentText = sc.comment
        self.selectedCodeBlocks = []
        self.imageURLs = []
        self.videoURLs = []
        self.audioURLs = []
        self.linesAbove = 2
        self.linesBelow = 2
        try:
            self.currentFilePath = evt.fileId
        except:
            self.currentFilePath = evt.directoryId
        self.viewableBlogText = ""
        self.commentTags = []
        if len(sc.answers) > 0:
            self.questionCommentData = {
                "allAnswers": sc.answers[:],
                "correctAnswer": sc.answers[sc.correct_answer],
                "question": sc.question
            }
        else:
            self.questionCommentData = {
                "allAnswers": [],
                "correctAnswer": "",
                "question": ""
            }

        Comment.comments.append(self)
        Comment.seq_num += 1

class Footnote:
    def __init__(self, value, line, col, highlights):
        self.value = int(normalize("NFKC", value))
        self.col = col
        self.line = line
        self.highlights = highlights[:]

class StoryComment:
    def __init__(self, comment, questions):
        self.comment = markdown.markdown(comment)
        self.question = ""
        self.answers = []
        self.correct_answer = 0

        if len(questions) > 0:
            self.question = markdown.markdown(questions[0]).replace("<p>", "").replace("</p>", "")
            for i, q in enumerate(questions[1:]):
                self.answers.append(markdown.markdown(q[0]).replace("<p>", "").replace("</p>", ""))
                if q[1]:
                    self.correct_answer = i

def walk(path):
    for p in Path(path).iterdir():
        if p.is_dir():
            yield [p.resolve(), True]
            yield from walk(p)
            continue
        yield [p.resolve(), False]

abs_path = len(str(Path(import_dir).resolve())) + 1

evt = CreateDirEvent("/")
evt.timestamp = 0

# Parse Markdown comments
story_comments = defaultdict(list)
questions = []

with open(import_dir + "/story.md") as f:
    out_str = ""

    def new_comment():
        global lineno
        global out_str
        global story_comments
        global questions

        if len(out_str) > 0:
            story_comments[lineno].append(StoryComment(out_str, questions))
            out_str = ""
        lineno = int(line[2:])
        questions = []

    for line in f:
        if line.startswith("# "):
            new_comment()
        elif line.startswith("## "):
            # Question
            questions = [line[2:].strip()]
            for line in f:
                if line.startswith("- "):
                    questions.append([line[2:].strip(), False])
                elif line.startswith("* "):
                    questions.append([line[2:].strip(), True])
                elif line.startswith("# "):
                    new_comment()
                    break
        else:
            out_str += line
    if len(out_str) > 0:
        story_comments[lineno].append(StoryComment(out_str, questions))

if 0 in story_comments:
    for sc in story_comments[0]:
        comment = Comment(evt, sc)

def insert_file(path, filename):
    col = 1
    line = 1
    with open(str(path), "r") as f:
        while len(c := f.read(1)) > 0:
            evt = InsertFileEvent(filename)
            evt.lineNumber = line
            evt.column = col

            if c == '\n':
                evt.character = "NEWLINE"
                line += 1
                col = 1
            else:
                evt.character = c
                col += 1

files = []

project = json.load(open(f"{import_dir}/story.json", "r"))

for d in project["directories"]:
    CreateDirEvent(d)
    print("Dir:", d)

for f in project["files"]:
    files.append(f)
    CreateFileEvent(f)
    insert_file(f"{import_dir}/{f}", f)
    print("File:", d)

diffs = {}

for d in project["diffs"]:
    f = d["file"]
    diffs[f] = d
    files.append(f)
    CreateFileEvent(f)
    insert_file(f"{import_dir}/{d['content']}", f)
    print("Diff File:", d)

Event.init_done = True

def add_comment(evt, l, c):
    for fn in footnotes:
        if (fn.line, fn.col) == (l, c):
            for sc in story_comments[fn.value]:
                comment = Comment(evt, sc)
                for hl in fn.highlights:
                    comment.selectedCodeBlocks.append({
                        "fileId": comment.currentFilePath,
                        "selectedText": "",
                        "startRow": hl[0][0] - 1,
                        "startColumn": hl[0][1] - 1,
                        "endRow": hl[1][0] - 1,
                        "endColumn": hl[1][1] - 1
                    })

            return True
    return False


for fil, step in project["steps"]:
    file_diff = diffs[fil]
    footnotes = []
    fn_highlights = []

    if step == 0:
        # First
        with open(f"{import_dir}/{file_diff['content']}", "r") as f:
            file1 = f.readlines()

        file_diff["prev_step"] = file_diff["steps"][step]
        with open(f"{import_dir}/{file_diff['prev_step']}", "r") as f:
            file2 = f.readlines()
    else:
        with open(f"{import_dir}/{file_diff['prev_step']}", "r") as f:
            file1 = f.readlines()

        file_diff["prev_step"] = file_diff["steps"][step]
        with open(f"{import_dir}/{file_diff['prev_step']}", "r") as f:
            file2 = f.readlines()

    # Process file2 and remember footnotes
    col = 1
    line = 1
    last_was_footnote = False
    for i, s in enumerate("".join(file2)):
        if s == "⁽":
            fn_highlights.append([(line, col),()])
        elif s == "⁾":
            fn_highlights[-1][1] = (line, col)
        elif s in footnote_ch:
            if last_was_footnote:
                footnotes[-1].value = footnotes[-1].value * 10 + (int(normalize("NFKC", s)))
            else:
                last_was_footnote = True
                f = Footnote(s, line, col, fn_highlights)
                footnotes.append(Footnote(s, line, col - 1, fn_highlights))
                fn_highlights = []
        else:
            last_was_footnote = False
            col += 1
        if s == '\n':
            col = 1
            line += 1
            last_was_footnote = False

    # Strip footnote characters
    for i, s in enumerate(file1):
        file1[i] = "".join(x for x in s if x not in footnote_ch_full)
    for i, s in enumerate(file2):
        file2[i] = "".join(x for x in s if x not in footnote_ch_full)

    #print(f"Footnotes: {footnotes}")

    diff = SequenceMatcher(None, file1, file2)

    col = 1
    line = 1

    for tag, i1, i2, j1, j2 in diff.get_opcodes():
        print('{:7}   a[{}:{}] --> b[{}:{}] {!r:>8} --> {!r}'.format(
            tag, i1, i2, j1, j2, file1[i1:i2], file2[j1:j2]))
        a = "".join(file1[i1:i2])
        b = "".join(file2[j1:j2])
        if tag == "replace":
            delete = []

            def process_delete():
                global delete
                global col
                global line
                global c

                if len(delete) == 0:
                    return

                if len(delete) <= 10:
                    # Simulate "del" key
                    for d in delete:
                        evt = DeleteFileEvent(fil)
                        evt.lineNumber = line
                        evt.column = col
                        evt.character = d
                        add_comment(evt, line, col)
                else:
                    # Simulate "backspace" key
                    for i, d in enumerate(delete):
                        delete[i] = (d, col)
                        col += 1

                    for d, dcol in delete[::-1]:
                        evt = DeleteFileEvent(fil)
                        evt.lineNumber = line
                        evt.column = dcol
                        evt.character = d
                        col = dcol
                        add_comment(evt, line, dcol)

                delete = []

            for d in ndiff(a, b):
                #print(d)
                op = d[0]
                c = d[2:]

                if op != '-' or c == '\n':
                    process_delete()

                if op == "+":
                    evt = InsertFileEvent(fil)
                    evt.lineNumber = line
                    evt.column = col
                    evt.character = c
                    if c == "\n":
                        evt.character = "NEWLINE"
                elif op == "-":
                    # Count how many - are there and then delete from right to left as this looks nicer
                    if c == '\n':
                        evt = DeleteFileEvent(fil)
                        evt.lineNumber = line
                        evt.column = col
                        evt.character = "NEWLINE"
                    else:
                        delete.append(c)
                    continue

                add_comment(evt, line, col)
                if c == f'\n':
                    line += 1
                    col = 1
                else:
                    col += 1

            process_delete()

        elif tag == "delete":
            delete = []

            def process_delete():
                global delete
                global col
                global line
                global c

                if len(delete) == 0:
                    return

                if len(delete) <= 10:
                    # Simulate "del" key
                    for d in delete:
                        evt = DeleteFileEvent(fil)
                        evt.lineNumber = line
                        evt.column = col
                        evt.character = d
                        add_comment(evt, line, col)
                else:
                    # Simulate "backspace" key
                    for i, d in enumerate(delete):
                        delete[i] = (d, col)
                        col += 1

                    for d, dcol in delete[::-1]:
                        evt = DeleteFileEvent(fil)
                        evt.lineNumber = line
                        evt.column = dcol
                        evt.character = d
                        col = dcol
                        add_comment(evt, line, dcol)

                delete = []

            for c in a:
                # Count how many - are there and then delete from right to left as this looks nicer
                if c == '\n':
                    process_delete()
                    evt = DeleteFileEvent(fil)
                    evt.lineNumber = line
                    evt.column = col
                    evt.character = "NEWLINE"
                else:
                    delete.append(c)
                continue

            process_delete()
            #raise ValueError("Delete unsupported :(")

        elif tag == "insert":
            # Remember everything until the linebreak, add the linebreak and then input the text
            # This looks more natural
            chars = []
            for c in b:
                if c != '\n':
                    chars.append(c)
                else:
                    evt = InsertFileEvent(fil)
                    evt.character = "NEWLINE"
                    evt.lineNumber = line
                    evt.column = col

                    for cc in chars:
                        evt = InsertFileEvent(fil)
                        evt.character = cc
                        evt.lineNumber = line
                        evt.column = col
                        add_comment(evt, line, col)

                        col += 1


                    col = 1
                    line += 1
                    chars = []
        elif tag == "equal":
            for c in b:
                for fn in footnotes:
                    if (fn.line, fn.col) == (line, col):
                        evt = NoopEvent(fil)
                        evt.lineNumber = line
                        evt.column = col
                        assert(add_comment(evt, line, col))

                if c == '\n':
                    line += 1
                    col = 1
                else:
                    col += 1

        print("---")

with open(f"{playback_dir}/playback/events_{output_name}.json", "w") as f:
    d = {"events": Event.events}
    json.dump(d, f, indent=2, default=lambda o: o.__dict__)

from itertools import groupby

with open(f"{playback_dir}/playback/comments_{output_name}.json", "w") as f:
    d = {}
    for k, g in groupby(Comment.comments, lambda x: x.displayCommentEvent.id):
        d[k] = list(g)

    d = {"comments": d}

    json.dump(d, f, indent=2, default=lambda o: o.__dict__)

