---
title: "Prozedurale Programmierung -- Funktionen & Projektstruktur"
author: "Gabriel Kind"
institute: "Hochschule Mittweida"
date: April 2022
---

# Funktionen

## Definition: Algorithmus (laut Wikipedia)

- Beschreibt eine allgemeine Handlungsvorschrift, zur Lösung eines Problems oder einer Problemklasse
- Besteht aus endlich vielen, wohldefinierten Einzelschritten
- Können in ein Computerprogramm implementiert werden
- Aus einer bestimmten Eingabe resultiert eine bestimmte Ausgabe

## Definition: Algorithmus (Visuell)

![](Images/IO.png)

## Definition

- Ein Algorithmus ist ineinander abgeschlossen
- In C++ verwendet man dafür Funktionen

## Funktionsaufbau

```cpp
Rueckgabetyp Funktionsname(Parameter1, Parameter2, ...) {
    /* Funktionsinhalt */
    return Rückgabewert
}
```

## Funktionsaufbau (Beispiel)

```cpp
/**
 * Zählt die Elemente in einem Vektor mit Wert 1
 * @param vec Vektor
 * @return Anzahl Elemente mit Wert 1
 */
int anzahl_wert1(const vector<int>& vec) {
    int anzahl = 0;
    for (int z: vec) {
        if (z == 1) {
            ++anzahl;
        }
    }
    return anzahl;
}
std::vector<int> zahlen = /* Werte hier */;
anzahl_wert1(zahlen);
```

## Funktionsaufbau (Anmerkungen)

Der ``/**/`` ist ein Dokumentationskommentar im javadoc-Format (Ursprünglich von Java).

Der Vektor wird als eine konstante (``const``) Referenz (``&``) übergeben. ``const`` bedeutet Modifikationen am Vektor sind innerhalb der Funktionen verboten. Ohne das ``&`` würde C++ den Vector elementweise kopieren. Dies ist ziemlich ineffizient, daher übergeben wir hier eine Referenz.

# Projektstruktur

## Projektstruktur

Bis jetzt haben wir nur mit einzelnen Dateien gearbeitet.

In C/C++ unterscheidet man streng zwischen _Deklaration_ und _Implementierung_.

- Deklaration von Klassen und Funktion: Im Header (Endung ".h")
- Implementierung der Funktionen (der eigentliche Code): In der Code-Datei (Endung ".cpp" oder ".cxx")

## Header-Datei

Beispiel für eine Datei ``mein_code.h``

```cpp
#pragma once // <-- Wird später erklärt

int add(int x, int y);
int add(int x, int y, int z);
```

Diese Zeilen heißen _Funktionssignatur_.

## Quellcode-Datei

Beispiel für eine Datei ``mein_code.cpp``.

Diese implementiert die Funktionen aus dem Header.

```cpp
#include "mein_code.h"

int add(int x, int y) {
    return x + y;
}

int add(int x, int y, int z) {
    return x + y + z;
}
```

## Verwendung der Funktionen aus ``mein_code``

```cpp
#include <iostream>
#include "mein_code.h"

int main() {
    std::cout << add(3, 4) << "\n";
    std::cout << add(4, 6, 8) << "\n";
    return 0;
}
```

## CMakeLists

In ``add_executable`` müssen alle cpp-Dateien stehen (Header sind egal).

Reihenfolge ist egal. Ich sortiere gerne alphabetisch.

```cmake
add_executable(mein_programm
    main.cpp
    mein_code.cpp
)
```

## Include-Guard

Aus historischen Gründen müssen Header-Dateien immer einen Include-Guard enthalten.

Dieser muss eindeutig sein!

Wenn dieser fehlt kann es zu merkwürdigen Compilerfehlern kommen.

```cpp
#ifndef MEIN_HEADER_H
#define MEIN_HEADER_H

/* Alle Deklarationen hier */

#endif // Ende: MEIN_HEADER_H
```

## Include-Guard Alternative

Nicht standardkonform, aber von quasi jedem Compiler unterstützt ist ``#pragma once`` anstatt eines Include-Guards:

```cpp
#pragma once

/* Alle Deklarationen hier */
```

## Vollständiger Header:

:::::: {.columns}
::: {.column}

```cpp
#ifndef MEIN_HEADER_H
#define MEIN_HEADER_H

int add(int x, int y);
int add(int x, int y, int z);

#endif
```

:::
::: {.column}

```cpp
#pragma once

int add(int x, int y);
int add(int x, int y, int z);
```

:::
::::::

## Welchen Include Guard verwenden?

- Geschmackssache
- Projektabhängig

# Zufallszahlen generieren

## Achtung: Veraltete Funktionen

Im Internet finden Sie jede Menge Code mit ``srand`` und ``rand``. 

Dies ist _veraltet_ und sollte nicht mehr verwendet werden.

Die Verwendung von ``rand`` ist fehleranfällig und Code aus dem Internet oft _falsch_!

Videoempfehlung: [rand() Considered Harmful](https://channel9.msdn.com/Events/GoingNative/2013/rand-Considered-Harmful)

## Empfohlener Code (Benötigt ``<random>``)

```cpp
std::random_device rd; // Zufallsgenerator erzeugen
std::mt19937 gen(rd()); // Den Generator seeden
// dist erzeugt bei Aufruf Zahlen in [1, 10]
std::uniform_int_distribution<> dist(1, 10);

int zahl = dist(gen); // erste Zufallszahl
int zahl2 = dist(gen); // zweite Zufallszahl usw...
```

## Seed

Computer sind deterministisch, d.h. es gibt keinen Zufall!

Der Zufall wird anhand von Funktionen mit sehr langen Perioden erzeugt, wodurch es zufällig wirkt.

Der Seed ist die Initialisierung für den Zufallszahlengenerator.

Gleicher Seed = Gleiche Zufallszahlen.

## Kryptografisch sichere Zufallszahlen

**Achtung**. ``mt19937`` und auch die anderen in C++ eingebauten Verfahren sind _nicht_ kryptografisch sicher, d.h. für die üblichen Fälle, wo man Zufall braucht, sind sie gut genug, aber _nicht_ für z.B. verschlüsselte Kommunikation.

# Danke für Ihre Aufmerksamkeit! Fragen?
