---
title: "Prozedurale Programmierung -- Speicherverwaltung"
author: "Gabriel Kind"
institute: "Hochschule Mittweida"
date: Juni 2022
---

## Stack und Heap (Wiederholung)

- Daten werden in C++ auf dem Stack bzw. dem Heap abgelegt
- Stackspeicher wird automatisch verwaltet
- Heapspeicher muss vom Betriebssystem angefordert und auch wieder freigegeben werden

## Der Stack in C++

- Auf dem Stack werden lokale Variablen und Rücksprungadressen gespeichert
- Die Rücksprungadresse ist der Ort, wo nach einem ``return`` fortgesetzt wird
- Bei jedem Funktionsaufruf wird ein _Stackframe_ erzeugt, bestehend aus:
	- Rücksprungadresse
	- Platz für lokale Variablen
- Hat begrenze Größe: Stapelüberlauf (_Stackoverflow_) wenn am Limit
- Ist systemabhängig und teilweise nur wenige KB groß: Große Speicheranforderungen müssen auf dem Heap passieren
- Wird automatisch freigegeben, wenn ein Scope verlassen wird

## Beispiele

- ``int bla``: Der Integer wird gelöscht, wenn der Scope verlassen wird
- ``std::vector<int>``: Der Vektor ist auf dem Stack, die Daten liegen aber auf dem Heap. Wenn der Scope verlassen wird, werden die Heap-Objekte mitgelöscht
- ``std::string``: Wie beim Vektor, nur das hier Zeichen auf dem Heap liegen
- Funktioniert alles vollautomatisch, nichts zu beachten
- siehe Beispiel 01 und Storyteller 01

## Datenstruktur: Stack

- Der Stack ist eine Datenstruktur ähnlich dem Array
- Operationen: push und pop
- Push: Füge Element dem Stack hinzu
- Pop: Entferne Element vom Stack

## Push/Pop illustriert

![(Originalquelle und Lizenz unklar)](Images/PushPop.jpg)

## Heap

- Heap wird vom Betriebssystem verwaltet
- Maximalgröße: Etwa der Adressraum (4 Gigabyte bei 32 Bit, 16 Exabyte bei 64 Bit)
- Anfordern mit ``new``, Rückgabe mit ``delete`` (Manuelle Verwaltung)
- siehe Beispiel 02 und Storyteller 02

## Rückblick

- Bis jetzt haben wir nur mit dem Stack gearbeitet
- Speicherverwaltung war also bis jetzt kein Thema
- Wir haben aber schon indirekt mit dem Heap gearbeitet, z.B. mit ``std::vector`` und ``std::string``

## Raw Pointer (Veraltet)

- Bereits von Iteratoren bekannt
- _Zeigen_ auf Speicher (enthalten Speicheradresse)
- Diese Adresse kann mit ``*`` dereferenziert werden

## Allokation

- Erzeugen: Mit ``new``:
	- ``Point* p = new Point(1, 2);``
- Speicher freigeben:
	- ``delete p``
- Wenn es keine Pointer mehr gibt, die auf den Speicher zeigen, ist ein Speicherleck (_Memory Leak_) entstanden. Dies ist ein **Programmierfehler**
- Zugriff auf freigegebenen Speicher mittels Pointer (_use-after-free_) ist eine **Sicherheitslücke**
- Freigeben von freigegebenem Speicher (_double-free_) ist eine **Sicherheitslücke**
- Siehe Beispiel 03 und Storyteller 03

## Beispiel

```cpp
Point* p = new Point(1, 2);
```

Wenn der Scope verlassen wird, wird nur der Pointer gelöscht, aber nicht das Heap-Objekt.

Es muss immer manuell ``delete p`` ausgeführt werden.

## Wann benötigt man Heap-Speicher

- Objekt ist zu groß für den Stack
- Objekt hat dynamische Größe
- Objekt wird viel zwischen Funktionen herumgereicht und Kopie/Verschiebung des Objektes ist nicht möglich oder langsam

## Probleme von Raw Pointern

- ``new`` gibt einen _Raw Pointer_ zurück
- Besitz unklar, muss dokumentiert werden, wer den Pointer löschen muss
- Wenn ein Objekt mehreren gehört (z. B. ein Bild wird mehrmals referenziert) muss sichergestellt werden, dass erst gelöscht wird, wenn alle Referenzen weg sind
- Kompliziert und fehleranfällig
- Die Lösung: Smart Pointer

# Smart Pointer

## Smart Pointer

Manuelle Speicherverwaltung mit ``new`` und ``delete`` ist seit C++11 veraltet.

Smart Pointer stellen sicher, dass ein Objekt gelöscht wird, sobald es nicht mehr benötigt wird.

Seit C++11 gibt es drei Klassen:

- ``std::unique_ptr`` für exklusiven Besitz
- ``std::shared_ptr`` für geteilten Besitz
- ``std::weak_ptr`` um Zyklen aufzulösen (zusammen mit ``std::shared_ptr``)

## Beispiele

Für ``std::unique_ptr``:

- siehe Beispiele 04, 05 und 06
- siehe Storyteller 03 und 04

Für ``std::shared_ptr`` und ``std::weak_ptr``:

- siehe Beispiele 07 und 08
- siehe Storyteller 05 und 06

## Vorteil gegenüber Raw Pointern

Smart Pointer liegen auf dem Stack, verwalten aber Heap-Speicher.

Wenn der Scope endet, wird der Heapspeicher freigegeben.

Funktioniert also wie bei ``std::string`` oder ``std::vector<int>``.

## ``std::auto_ptr``

Es gibt auch noch ``std::auto_ptr``, dieser ist veraltet und hat mehrere Designprobleme, wodurch er nicht richtig funktioniert.

Wurde in C++17 entfernt.

**Nicht verwenden**

Der Nachfolger ist ``std::unique_ptr``.

## ``std::unique_ptr``

- Für exklusiven Besitz
- Objekte im ``std::unique_ptr`` können nicht kopiert werden
- Der Besitz ist also immer eindeutig geregelt
- Löscht das Objekt, wenn der Scope verlassen wird
- Mit ``get()`` können Raw-Pointer erhalten werden (nicht mit ``delete`` löschen, das Objekt gehört Ihnen nicht!)

## ``std::unique_ptr``: Besitzübertragung

- Der von ``unique_ptr`` verwaltete Speicher kann an einen anderen ``unique_ptr`` abgegeben werden.
- Dafür verwendet man ``std::move``.

## Verschiebung mit ``std::move``

``std::move`` wird immer verwendet, um den Besitz abzugeben.

Zudem wird es genutzt, wenn das Objekt nicht kopierbar ist, dies trifft auf Dateistreams zu.

## ``std::shared_ptr`` und ``std::weak_ptr``

Limitierung von ``std::unique_ptr``: Exklusiver Besitz

Wenn nicht klar ist, ob das Objekt gelöscht werden darf (komplizierter Besitz), empfiehlt sich ein ``std::shared_ptr``.

Dieser Pointer zählt, wie oft er verwendet wird (Referenzzähler) und löscht sich, wenn es keine Nutzer mehr gibt.

Problem: Zyklen. Diese können mit ``std::weak_ptr`` aufgebrochen werden.

# Danke für Ihre Aufmerksamkeit! Fragen?
