#include "engine.h"

#include <cmath>
#include <iostream>

Engine::Engine(SDL_Renderer* renderer, int window_width, int window_height) {
    this->renderer = renderer;
    this->window_width = window_width;
    this->window_height = window_height;

    rect = Rectangle(Point(100, 100), 20, 20);
    rect.set_color(255, 0, 0);
    rect.velocity = Point(1, 1);

    cursor = Circle(Point(-100, -100), 5);
    cursor.set_color(0, 255, 255);
}

bool Engine::process_events() {
    SDL_Event evt;

    while (SDL_PollEvent(&evt)) {
        // Jeweils ein Event verarbeiten
        if (evt.type == SDL_QUIT) {
            // Benutzer hat Programmende angefordert (Fenster geschlossen)
            return false;
        } else if (evt.type == SDL_KEYDOWN) {
            // Tastatur-Taste gedrückt
            if (evt.key.keysym.scancode == SDL_SCANCODE_W) {
                // WSAD
                // W oben
                // S unten
                // A links
                // D rechts
                rect.velocity.y -= 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_A) {
                rect.velocity.x -= 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_S) {
                rect.velocity.y += 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_D) {
                rect.velocity.x += 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_LSHIFT) {
                // Bremsen, wenn Linke Umschalttaste gedrückt wird
                do_break = true;
            }
        } else if (evt.type == SDL_KEYUP) {
            if (evt.key.keysym.scancode == SDL_SCANCODE_LSHIFT) {
                // Bremsen beenden, wenn Linke Umschalttaste losgelassen
                do_break = false;
            }
        } else if (evt.type == SDL_MOUSEMOTION) {
            // Mausbewegung im Fenster
            int x = evt.motion.x;
            int y = evt.motion.y;

            Point& p = cursor.position;
            p.move_to(x - cursor.get_radius(), y - cursor.get_radius());
        } else if (evt.type == SDL_MOUSEWHEEL) {
            // Mausrad
            // Das Mausrad ändert den Grauton des Hintergrunds
            // Achtung: Muss im Bereich 0 bis 255 bleiben!
            if (evt.wheel.y > 0) {
                background_color += 10;

                if (background_color > 255) {
                    background_color = 255;
                }
            } else if(evt.wheel.y < 0) {
                background_color -= 10;

                if (background_color < 0) {
                    background_color = 0;
                }
            }
        } else if (evt.type == SDL_MOUSEBUTTONDOWN) {
            // Gegenteil: SDL_MOUSEBUTTONUP
            // Maustaste gedrückt
            if (evt.button.button == SDL_BUTTON_LEFT) {
                // Linke Maustaste gedrückt
                int x = evt.button.x;
                int y = evt.button.y;

                // Abstand zwischen Mausklick und Rechteck
                x = x - rect.position.x;
                y = y - rect.position.y;

                // Wenn geklickt wird, soll das Rechteck zum Klickort wandern
                double angle = atan2(y, x);
                rect.velocity.x = cos(angle) * 2;
                rect.velocity.y = sin(angle) * 2;
            } else if (evt.button.button == SDL_BUTTON_RIGHT) {
                // Linke Maustaste gedrückt
                int x = evt.button.x;
                int y = evt.button.y;

                // Abstand zwischen Mausklick und Rechteck
                x = rect.position.x - x;
                y = rect.position.y - y;

                // Wenn geklickt wird, soll das Rechteck zum Klickort wandern
                double angle = atan2(y, x);
                rect.velocity.x = cos(angle) * 10;
                rect.velocity.y = sin(angle) * 10;
            } else if (evt.type == SDL_CONTROLLERDEVICEADDED) {
                // Controller wurde angesteckt
                int id = evt.cdevice.which;
                if (SDL_IsGameController(id)) {
                    SDL_GameController* controller = SDL_GameControllerOpen(id);
                    if (controller) {
                        std::cout << "Controller added: " << SDL_GameControllerName(controller) << "\n";
                    }
                }
            } else if (evt.type == SDL_CONTROLLERDEVICEREMOVED) {
                int id = evt.cdevice.which;
                SDL_GameController* controller = SDL_GameControllerFromInstanceID(id);
                if (controller) {
                    std::cout << "Controller removed: " << SDL_GameControllerName(controller) << "\n";
                    SDL_GameControllerClose(controller);
                }
            } else if (evt.type == SDL_CONTROLLERBUTTONDOWN) {
                // Gegenteil: SDL_CONTROLLERBUTTONUP
                if (evt.cbutton.button == SDL_CONTROLLER_BUTTON_A) {
                    if (rect.green == 0) {
                        rect.green = 255;
                    } else {
                        rect.green = 0;
                    }
                }
            }
        }
    }

    return true;
}

void Engine::update_and_render() {
    // Fenster schwarz ausmalen
    SDL_SetRenderDrawColor(renderer, background_color, background_color, background_color, 0xFF);
    SDL_RenderClear(renderer);

    /* Eigener Code hier */
    rect.update();
    rect.draw(renderer);

    cursor.draw(renderer);

    if (rect.position.x + rect.width > window_width || rect.position.x < 0) {
        rect.velocity.x = -rect.velocity.x;
    }
    if (rect.position.y + rect.height > window_height || rect.position.y < 0) {
        rect.velocity.y = -rect.velocity.y;
    }

    if (do_break) {
        if (rect.velocity.x < 0) {
            rect.velocity.x += 0.1;
        } else if (rect.velocity.x > 0) {
            rect.velocity.x -= 0.1;
        }

        if (rect.velocity.y < 0) {
            rect.velocity.y += 0.1;
        } else {
            rect.velocity.y -= 0.1;
        }

        // -0.05 + 0.1 -> 0.05
        // Damit bleibt das Rechteck wirklich stehen
        if (abs(rect.velocity.x) <= 0.1) {
            rect.velocity.x = 0.0;
        }

        if (abs(rect.velocity.y) <= 0.1) {
            rect.velocity.y = 0.0;
        }
    }

    // Aenderungen sichtbar machen
    SDL_RenderPresent(renderer);
}
