#include "engine.h"
#include "rectangle.h"

#include <SDL_events.h>
#include <cmath>
#include <iostream>

Engine::Engine(SDL_Renderer* renderer, int window_width, int window_height) {
    this->renderer = renderer;
    this->window_width = window_width;
    this->window_height = window_height;

    rect = Rectangle(Point(100, 100), 20, 20);
    rect.set_color(255, 0, 0);
    rect.velocity = Point(0, 0);

    cursor = Circle(Point(-100, -100), 5);
    cursor.set_color(255, 0, 0);
}

bool Engine::process_events() {
    // Auf Nutzereingaben reagieren
    // siehe https://wiki.libsdl.org/SDL_Event
    SDL_Event evt;

    while (SDL_PollEvent(&evt)) {
        // Jeweils ein Event verarbeiten
        if (evt.type == SDL_QUIT) {
        // Benutzer hat Programmende angefordert (Fenster geschlossen)
            return 0;
        } else if (evt.type == SDL_KEYDOWN) {
            // Tastatur-Taste gedrückt
            if (evt.key.keysym.scancode == SDL_SCANCODE_A) {
                rect.velocity.x -= 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_S) {
                rect.velocity.y += 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_W) {
                rect.velocity.y -= 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_D) {
                rect.velocity.x += 1;
            } else if (evt.key.keysym.scancode == SDL_SCANCODE_SPACE) {
                do_break = true;
            }
        } else if (evt.type == SDL_KEYUP) {
            // Tastatur-Taste losgelassen
            if (evt.key.keysym.scancode == SDL_SCANCODE_SPACE) {
                do_break = false;
            }
        } else if (evt.type == SDL_MOUSEBUTTONDOWN) {
            // Maustaste gedrückt
            if (evt.button.button == 1) {
                int x = evt.button.x - (rect.position.x + rect.width / 2);
                int y = evt.button.y - (rect.position.y + rect.height / 2);
                auto angle = atan2(y, x);
                rect.velocity.x = cos(angle) * 2;
                rect.velocity.y = sin(angle) * 2;
            }
        } else if (evt.type == SDL_MOUSEBUTTONUP) {
            // Maustaste losgelassen
        } else if (evt.type == SDL_MOUSEMOTION) {
            // Maus im Fenster bewegt
            Point& c = cursor.position;
            c.move_to(evt.motion.x - cursor.get_radius(), evt.motion.y - cursor.get_radius());

            // Idee: Der Cursor ist im Rechteck, wenn einer der Eckpunkte des Rechtecks
            cursor.is_inside(rect.position);
        } else if (evt.type == SDL_MOUSEWHEEL) {
            if (evt.wheel.y > 0) {
                // Up
                background_color += 10;
                if (background_color > 255) {
                    background_color = 255;
                }
            } else if (evt.wheel.y < 0) {
                // Down
                background_color -= 10;
                if (background_color < 0) {
                    background_color = 0;
                }
            }
        } else if (evt.type == SDL_CONTROLLERDEVICEADDED) {
            int id = evt.cdevice.which;
            if (SDL_IsGameController(id)) {
                SDL_GameController* controller = SDL_GameControllerOpen(id);
                if (controller) {
                    std::cout << "Controller added: " << SDL_GameControllerName(controller) << "\n";
                }
            }
        } else if (evt.type == SDL_CONTROLLERDEVICEREMOVED) {
            int id = evt.cdevice.which;
            SDL_GameController* controller = SDL_GameControllerFromInstanceID(id);
            if (controller) {
                std::cout << "Controller removed: " << SDL_GameControllerName(controller) << "\n";
                SDL_GameControllerClose(controller);
            }
        } else if (evt.type == SDL_CONTROLLERBUTTONDOWN) {
            if (evt.cbutton.button == SDL_CONTROLLER_BUTTON_A) {
                
            }
        }
    }

    return true;
}

void Engine::update_and_render() {
    // Fenster schwarz ausmalen
    SDL_SetRenderDrawColor(renderer, background_color, background_color, background_color, 0xFF);
    SDL_RenderClear(renderer);

    /* Eigener Code hier */
    /*SDL_Rect rect = { 0, 0, 100, 50 };
    SDL_SetRenderDrawColor(renderer, 0xFF, 0x0, 0x0, 0xFF);
    SDL_RenderFillRect(renderer, &rect);*/

    if (do_break) {
        if (abs(rect.velocity.x) < 0.1) {
            rect.velocity.x = 0.0;
        } else {
            // Nimmt Wert des 1. Parameters und das Vorzeichen des 2. Parameters
            // Bei positiver Beschleunigung wird also 0.1 abgezogen, bei negativer wird 0.1 addiert
            rect.velocity.x -= std::copysign(0.1, rect.velocity.x);
        }

        if (abs(rect.velocity.y) < 0.1) {
            rect.velocity.y = 0.0;
        } else {
            rect.velocity.y -= std::copysign(0.1, rect.velocity.y);
        }
    }

    rect.update();
    rect.draw(renderer);

    if (rect.position.x + rect.width > window_width || rect.position.x < 0) {
        rect.velocity.x = -rect.velocity.x;
    }

    if (rect.position.y + rect.height > window_height || rect.position.y < 0) {
        rect.velocity.y = -rect.velocity.y;
    }

    cursor.draw(renderer);

    // Aenderungen sichtbar machen
    SDL_RenderPresent(renderer);
}
